//
// ********************************************************************
// * DISCLAIMER                                                       *
// *                                                                  *
// * The following disclaimer summarizes all the specific disclaimers *
// * of contributors to this software. The specific disclaimers,which *
// * govern, are listed with their locations in:                      *
// *   http://cern.ch/geant4/license                                  *
// *                                                                  *
// * Neither the authors of this software system, nor their employing *
// * institutes,nor the agencies providing financial support for this *
// * work  make  any representation or  warranty, express or implied, *
// * regarding  this  software system or assume any liability for its *
// * use.                                                             *
// *                                                                  *
// * This  code  implementation is the  intellectual property  of the *
// * GEANT4 collaboration.                                            *
// * By copying,  distributing  or modifying the Program (or any work *
// * based  on  the Program)  you indicate  your  acceptance of  this *
// * statement, and all its terms.                                    *
// ********************************************************************
//
//
// $Id: G4PAIclusterModel.cc,v 1.2 2004/11/10 10:03:12 grichine Exp $
// GEANT4 tag $Name: geant4-07-00-patch-01 $
//

#include "G4Timer.hh"

#include "G4PAIclusterModel.hh"
#include "Randomize.hh"
#include "G4Material.hh"
#include "G4MaterialTable.hh"
#include "globals.hh"

#include "G4ParticleDefinition.hh"
#include "G4Electron.hh"
#include "G4Positron.hh"
#include "G4Gamma.hh"
#include "G4TransportationManager.hh"
#include "G4VSensitiveDetector.hh"
#include "G4VTouchable.hh"


////////////////////////////////////////////////////////////////////////////
//
// Constructor, destructor

G4PAIclusterModel::G4PAIclusterModel(G4Envelope *anEnvelope) :
  G4VClusterModel("G4PAIclusterModel",anEnvelope)
{
  const G4ParticleDefinition* theElectron = G4Electron::Electron();
  G4DataVector data;
  fPAIonisation = new G4PAIModel(theElectron,"PAIclusterModel");
  fPAIonisation->Initialise(theElectron,data);
}

///////////////////////////////////////////////////////////////////////////

G4PAIclusterModel::~G4PAIclusterModel()
{
  delete fPAIonisation ;
}

///////////////////////////////////////////////////////////////////////////////
//
// Returns condition for application of the model depending on particle type


G4bool G4PAIclusterModel::IsApplicable(const G4ParticleDefinition& particle)
{
  return  ( particle.GetPDGCharge() != 0.0 && particle.GetPDGMass() != 0 ) ; 
}

/////////////////////////////////////////////////////////////////////
//
// UserTrigger() method: method which has to decide if
// the parameterisation has to be applied.
// Here ModelTrigger() asks the user (ie you) a 0/1 answer.
//
// Note that quantities like the local/global position/direction etc..
// are available at this level via the fastTrack parameter (allowing 
// to check distance from boundaries, see below  to allow the decision)
//

G4bool G4PAIclusterModel::ModelTrigger(const G4FastTrack& fastTrack) 
{
  G4double kinEnergy, mass, gamma ;
  kinEnergy  = fastTrack.GetPrimaryTrack()->GetKineticEnergy() ;
  mass       = fastTrack.GetPrimaryTrack()->GetDefinition()->GetPDGMass() ;
  gamma      = 1.0 + kinEnergy/mass ;
  if (gamma >= 1.2) return true  ;
  else              return false ;
}

//////////////////////////////////////////////////////////////////////////////
//
// 

void G4PAIclusterModel::DoIt( const G4FastTrack& fastTrack , 
		                    G4FastStep&  fastStep         )
{
  G4double charge, charge2, kinEnergy, mass, massRatio, scaledTkin ;
  G4double distance, energyTransfer, lambda, step, stepSum = 0.0 ;
  G4double energyLoss = 0.0 ;
  G4ThreeVector clusterPosition ;

  fClusterPositionVector.clear() ;
  fClusterEnergyVector.clear() ;

  charge     = fastTrack.GetPrimaryTrack()->GetDefinition()->GetPDGCharge() ;
  charge2    = charge*charge ;
  kinEnergy  = fastTrack.GetPrimaryTrack()->GetKineticEnergy() ;
  mass       = fastTrack.GetPrimaryTrack()->GetDefinition()->GetPDGMass() ;
  massRatio  = proton_mass_c2/mass ;
  scaledTkin = kinEnergy*massRatio ;

  G4ParticleMomentum direction(fastTrack.GetPrimaryTrackLocalDirection());

  distance = fastTrack.GetEnvelopeSolid()->
             DistanceToOut(fastTrack.GetPrimaryTrackLocalPosition(),direction) ;

  G4ThreeVector position = fastTrack.GetPrimaryTrackLocalPosition() + 
                           distance*direction ;

  // Set final position:

  fastStep.SetPrimaryTrackFinalPosition(position);

  // Cluster counting loop

  lambda = 1*mm;// fPAIonisation->GetFreePath( scaledTkin, charge2 ) ;
  step   = RandExponential::shoot(lambda) ;
  //  if (step < 0.0) step = 0.0 ;
  stepSum += step ;
  //  distance -= stepSum ;
  //  if(distance < 0.0) // no change, return 
  if(stepSum > distance ) // no change, return 
  {
    return ;  
  }
  else
  {
    G4ThreeVector globalStartPosition  = fastTrack.GetPrimaryTrack()->
                                         GetPosition() ;
    G4ParticleMomentum globalDirection = fastTrack.GetPrimaryTrack()->
                                         GetMomentumDirection() ; 

    //  while(distance >= 0.0)  
    while(stepSum <= distance )  // global (or local ?) cluster coordinates
    {
      //  clusterPosition = fastTrack.GetPrimaryTrackLocalPosition() + 
      //                  stepSum*direction ;  
  
      clusterPosition = globalStartPosition  + stepSum*globalDirection ;    
      energyTransfer  = fPAIonisation->GetPostStepTransfer(scaledTkin) ;

      fClusterPositionVector.push_back(clusterPosition) ;      
      fClusterEnergyVector.push_back(energyTransfer) ;

      step = RandExponential::shoot(lambda) ;
      // if (step < 0.0) step = 0.0 ;
      stepSum    += step ;
      //  distance   -= step ;
      energyLoss += energyTransfer ;     
    } 
    kinEnergy -= energyLoss ;
    fastStep.SetPrimaryTrackFinalKineticEnergy(kinEnergy) ;  
    // fastStep.SetTotalEnergyDeposited(energyLoss);

    BuildDetectorResponse() ; 
  }
  return ;
}


//
//
///////////////////////////////////////////////////////////////////////












